"""MX-28 servo usage."""
from pydxl.dynamixel import Dynamixel
from pydxl.register import Register


class Mx28(Dynamixel):
    """Class for MX-28 servo."""

    # http://emanual.robotis.com/docs/en/dxl/mx/
    # mx-28/#control-table-of-eeprom-area
    model_number = Register(address=0, size=2)
    firmware_version = Register(address=2)
    id_ = Register(address=3, write=True)
    baud_rate = Register(address=4, write=True)
    return_delay_time = Register(address=5, write=True)
    cw_angle_limit = Register(address=6, write=True, size=2)
    ccw_angle_limit = Register(address=8, write=True, size=2)
    temperature_limit = Register(address=11, write=True)
    min_voltage_limit = Register(address=12, write=True)
    max_voltage_limit = Register(address=13, write=True)
    max_torque = Register(address=14, write=True, size=2)
    status_return_level = Register(address=16, write=True)
    alarm_led = Register(address=17, write=True)
    shutdown = Register(address=18, write=True)
    multi_turn_offset = Register(address=20, write=True, size=2)
    resolution_divider = Register(address=22, write=True)

    # http://emanual.robotis.com/docs/en/dxl/mx/
    # mx-28/#control-table-of-ram-area
    torque_enable = Register(address=24, write=True)
    led = Register(address=25, write=True)
    d_gain = Register(address=26, write=True)
    i_gain = Register(address=27, write=True)
    p_gain = Register(address=28, write=True)
    goal_position = Register(address=30, write=True, size=2)
    moving_speed = Register(address=32, write=True, size=2)
    torque_limit = Register(address=34, write=True, size=2)
    present_position = Register(address=36, size=2)
    present_speed = Register(address=38, size=2)
    present_load = Register(address=40, size=2)
    present_voltage = Register(address=42)
    present_temperature = Register(address=43)
    registered = Register(address=44)
    moving = Register(address=46)
    lock = Register(address=47, write=True)
    punch = Register(address=48, write=True, size=2)
    realtime_tick = Register(address=50, size=2)
    goal_acceleration = Register(address=73, write=True)



import time

import serial

from pydxl import DynamixelCommunicationException, DynamixelFailedOpeningPort


class PortHandler:
    def __init__(self, device="/dev/ttyUSB0", baudrate=1_000_000):
        self.device = device
        self.serial = None
        self._baudrate = baudrate
        self.open()
        self.packet_start_time = self._time()
        self.packet_timeout = 0

    def close(self):
        self.serial.close()

    def write(self, packet):
        nb_written_bytes = self.serial.write(packet.data)
        self.serial.flush()
        return nb_written_bytes

    def read(self, length):
        if self._time() >= self.packet_start_time + self.packet_timeout:
            raise DynamixelCommunicationException("timeout")
        return self.serial.read(length)

    @property
    def baudrate(self):
        return self._baudrate

    @baudrate.setter
    def baudrate(self, value):
        if value == self._baudrate:
            return  # already set to this baudrate
        if value not in (
            9600,
            19200,
            38400,
            57600,
            115_200,
            230_400,
            460_800,
            500_000,
            576_000,
            921_600,
            1_000_000,
            1_152_000,
            2_000_000,
            2_500_000,
            3_000_000,
            3_500_000,
            4_000_000,
        ):
            raise DynamixelFailedOpeningPort(f"Invalid baudrate: {value} bps")

        self.close()
        self.open()

    def open(self):
        self.serial = serial.Serial(
            port=self.device,
            baudrate=self.baudrate,
            bytesize=serial.EIGHTBITS,
            timeout=0,
        )
        if not self.serial.is_open:
            raise DynamixelFailedOpeningPort(f"Failed opening {device}")

    @staticmethod
    def _time():
        """Return time in milliseconds."""
        return round(time.time() * 1_000_000_000) / 1_000_000.0

    def set_timeout_for_length(self, length):
        tx_time_per_byte = (1000.0 / self.baudrate) * 10000.0
        self.packet_timeout = tx_time_per_byte * length + 34.0
        self.packet_start_time = self._time()



from textwrap import dedent

from pydxl import DynamixelMalFormedPacket


class Packet:
    def __init__(self, length):
        if length <= 4 or length > 250:
            raise DynamixelMalFormedPacket(
                f"Not allowed size of packet: {length}"
            )
        self.data = [0] * length
        self.data[0] = 255
        self.data[1] = 255
        self.data[3] = length - 4

    @classmethod
    def from_data(cls, data):
        packet = cls(len(data))
        packet.data = data
        if len(data) == packet.total_length:
            if not packet._check_checksum():
                raise DynamixelMalFormedPacket("Invalid checksum")
        return packet

    def __repr__(self):
        return dedent(
            f"""
            Packet
            | ID: {self.identifier}
            | instruction: {self.instruction}
            | paramater0: {self.parameter0}
            | payload: {self.payload}
            | length: {self.length}
            | error: {self.error}
            | raw_data: {self.data}
            """
        )

    @property
    def identifier(self):
        return self.data[2]

    @identifier.setter
    def identifier(self, value):
        self.data[2] = value

    @property
    def length(self):
        return self.data[3]

    @property
    def error(self):
        return self.data[4]

    @property
    def total_length(self):
        return self.length + 4

    @property
    def instruction(self):
        return self.data[4]

    @instruction.setter
    def instruction(self, value):
        self.data[4] = value

    @property
    def parameter0(self):
        return self.data[5]

    @parameter0.setter
    def parameter0(self, value):
        self.data[5] = value

    @property
    def parameter1(self):
        return self.data[6]

    @parameter1.setter
    def parameter1(self, value):
        self.data[6] = value

    @property
    def payload(self):
        return self.data[6 : self.total_length - 1]

    @payload.setter
    def payload(self, data):
        for index, value in enumerate(data):
            self.data[6 + index] = value

    def _compute_checksum(self):
        checksum = 0
        # compute checksum of everything except header, checksum
        for word in self.data[2 : self.total_length - 1]:
            checksum += word
        return ~checksum & 0xFF

    def write_checksum(self):
        self.data[self.total_length - 1] = self._compute_checksum()

    def _check_checksum(self):
        return self.data[self.total_length - 1] == self._compute_checksum()



import time

from pydxl import DynamixelCommunicationException, Packet

BROADCAST_ID = 254

# TODO
INST_READ = 2
INST_REG_WRITE = 4
INST_ACTION = 5
INST_FACTORY_RESET = 6
INST_CLEAR = 16


class PacketHandler:
    def __init__(self, port):
        self.port = port

    def write(self, packet):
        """Write a packet with the port handler."""
        packet.write_checksum()
        nb_written_bytes = self.port.write(packet)
        if nb_written_bytes != packet.total_length:
            raise DynamixelCommunicationException(
                "Problem in the number of sent bytes "
                f"({nb_written_bytes} out of {packet.total_length})"
            )

    def read(self):
        """Read a packet with the port handler."""
        # TODO check error code in packet.error
        data = []
        length_to_read = 6  # minimum packet size
        while True:
            data.extend(self.port.read(length_to_read - len(data)))

            header_pos = len(data)
            for i in range(len(data) - 1):
                if data[i] == 255 and data[i + 1] == 255:
                    # found header
                    header_pos = i
                    break
            else:
                continue  # header not found read more data
            if header_pos:
                data = data[header_pos:]

            packet = Packet.from_data(data)
            if packet.identifier >= BROADCAST_ID or packet.error > 127:
                del data[0]  # invalid packet, continue the reception
                continue

            length_to_read = packet.total_length
            if length_to_read == len(data):
                return packet

            time.sleep(0.03)

    def wait_ack(self, identifier):
        while True:
            self.port.set_timeout_for_length(8)
            response = self.read()
            if identifier == response.identifier:
                return response

    def write_bytes(self, identifier, address, data, size):
        data = int(data)
        if size == 1:
            data_write = [data]
        elif size == 2:
            data_write = [data & 0xFF, (data >> 8) & 0xFF]
        elif size == 4:
            data_write = [
                data & 0xFF,
                (data >> 8) & 0xFF,
                (data >> 16) & 0xFF,
                (data >> 24) & 0xFF,
            ]
        else:
            raise DynamixelCommunicationException("Unsupported size of data")

        packet = Packet(length=7 + len(data_write))
        packet.identifier = identifier
        packet.instruction = 3
        packet.parameter0 = address
        packet.payload = data_write
        self.write(packet)
        self.wait_ack(identifier)

    def read_bytes(self, identifier, address, size):
        packet = Packet(length=8)
        packet.identifier = identifier
        packet.instruction = 2
        packet.parameter0 = address
        packet.parameter1 = size
        self.write(packet)
        response = self.wait_ack(identifier)
        if size == 1:
            data = int(response.data[5])
        elif size == 2:
            data = int(response.data[5]) + (int(response.data[6]) << 8)
        elif size == 4:
            data = (
                int(response.data[5])
                + (int(response.data[6]) << 8)
                + (int(response.data[7]) << 16)
                + (int(response.data[8]) << 24)
            )
        return data

    def ping(self, identifier):
        if identifier >= BROADCAST_ID:
            raise DynamixelCommunicationException(
                f"Invalid identifier: {identifier}"
            )

        packet = Packet(length=6)
        packet.identifier = identifier
        packet.instruction = 1
        self.write(packet)
        self.wait_ack(identifier)



from pydxl import (
    DynamixelCommunicationException,
    DynamixelFailedOpeningPort,
    PortHandler,
)
from pydxl.protocol1 import PacketHandler as PacketHandler1

# TODO from dynamixel.protocol2 import PacketHandler as PacketHandler2


class SerialLink:
    def __init__(
        self, device="/dev/ttyUSB0", baudrate=1_000_000, protocol_version=1.0
    ):
        self.port_handler = PortHandler(device, baudrate)
        if int(10 * protocol_version) == 10:
            self.packet_handler = PacketHandler1(self.port_handler)
        else:
            self.packet_handler = PacketHandler2(self.port_handler)

    def close(self):
        self.port_handler.close()

    def ping(self, identifier):
        self.packet_handler.ping(identifier)

    def write_bytes(self, identifier, address, data, size):
        self.packet_handler.write_bytes(identifier, address, data, size)

    def read_bytes(self, identifier, address, size):
        return self.packet_handler.read_bytes(identifier, address, size)



"""Descriptor for a register."""


class Register:
    """Descriptor for a register."""

    def __init__(self, address=0, read=True, write=False, size=1):
        """Initializer."""
        self.address = address
        self.read = read
        self.write = write
        self.size = size

    def __repr__(self):
        """Represent a string."""
        if self.write:
            if self.read:
                rw = "R/W"
            else:
                rw = "W"
        else:
            rw = "R"
        return f"Reg(@{self.address}, {rw}, size: {self.size})"

    def __get__(self, instance, owner):
        """Getter."""
        # instance.attr
        if instance:
            return instance.read(self.address, self.size)
        # class.attr
        else:
            return self

    def __set__(self, instance, value):
        """Setter."""
        if self.write:
            instance.write(self.address, value, self.size)

    def __delete__(self, instance):
        """Deleter. Does nothing."""
        pass



"""Top-level package for Python Dynamixel."""

__author__ = """Vincent Poulailleau"""
__email__ = "vpoulailleau@gmail.com"
__version__ = "2019.03.19"

from pydxl.exceptions import (
    DynamixelCommunicationException,
    DynamixelFailedOpeningPort,
    DynamixelMalFormedPacket,
)
from pydxl.mx28 import Mx28
from pydxl.packet import Packet
from pydxl.port_handler import PortHandler
from pydxl.seriallink import SerialLink



class Dynamixel:
    def __init__(self, identifier, serial_link):
        self.identifier = identifier
        self.serial_link = serial_link

    @property
    def address_map(self):
        return {}

    def ping(self):
        return self.serial_link.ping(self.identifier)

    def write(self, address, value, size):
        self.serial_link.write_bytes(self.identifier, address, value, size)

    def read(self, address, size):
        return self.serial_link.read_bytes(self.identifier, address, size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = ["pyserial==3.4"]

setup_requirements = ["pytest-runner==4.4"]

test_requirements = ["pytest==4.3.1"]

setup(
    author="Vincent Poulailleau",
    author_email="vpoulailleau@gmail.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="Python package to use Dynamixel servos through a serial interface.",
    install_requires=requirements,
    license="BSD license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords="pydxl",
    name="pydxl",
    packages=find_packages(include=["pydxl*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/vpoulailleau/pydxl",
    version="2019.03.19",
    zip_safe=False,
)



class DynamixelException(Exception):
    pass


class DynamixelFailedOpeningPort(DynamixelException):
    pass


class DynamixelFailedSettingBaudrate(DynamixelException):
    pass


class DynamixelCommunicationException(DynamixelException):
    pass


class DynamixelMalFormedPacket(DynamixelException):
    pass
